/*	
 *	clipgym.c
 *
 *	P/ECE GYM Driver
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2004 Naoyuki Sawa
 *
 *	* Thu Dec 9 18:16:00 JST 2004 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/****************************************************************************
 *	GYMhCo֐
 ****************************************************************************/

static SN76489 gym_sn76489;

int
gym_init(GYMDRIVER* driver, const void* romp, int size)
{
	const GYMHEADER* header = romp;

	/* ܂NA܂B */
	memset(driver, 0, sizeof(GYMDRIVER));

	/* YM2612 chip܂B */
	{
		extern YM2612 gym_ym2612; /* RAMɔzu */
		driver->ym2612 = &gym_ym2612;
	}
	ym2612_init(driver->ym2612, 53693100/*NTSCMegaDrivěUNbN*/ / 7/*7YM2612֋*/);

	/* SN76489 chip܂B */
	driver->sn76489 = &gym_sn76489;
	sn76489_reset(driver->sn76489, 53693100/*NTSCMegaDrivěUNbN*/ / 15/*15SN76489֋*/);

	/* f[^TCYȂ܂B */
	if(size < sizeof(GYMHEADER)) {
		return -1;
	}

	/* VOl`܂B */
	if(memcmp(header->identifier, "GYMX", 4) != 0) {
		return -1;
	}

	/* k`ɂ͖ΉłB */
	if(header->uncompressed_size) {
		return -1;
	}

	/* raw gym dataʒu܂B */
	driver->data_pos = (const unsigned char*)(header + 1);

	/* raw gym dataI[ʒuݒ肵܂B */
	driver->data_end = driver->data_pos + (size - sizeof(GYMHEADER));

	/* YM2612 chip܂B */
	ym2612_init(driver->ym2612, 53693100/*NTSCMegaDrivěUNbN*/ / 7/*7YM2612֋*/);

	return 0; /*  */
}

int
gym_mix(GYMDRIVER* driver, short wbuff[/*GYMBUFLEN*/])
{
	static short dacbuf[GYMBUFLEN_DAC];
	int daccnt = 0;
	//
	YM2612* const ym2612 = driver->ym2612;
	SN76489* const sn76489 = driver->sn76489;
	//
	const unsigned char* data_pos = driver->data_pos; /* o */
	const unsigned char* data_end = driver->data_end; /* o */
	int tmpcnt = driver->tmpcnt; /* o */
	//
	int op;
	int reg;
	int data;

	while(tmpcnt < GYMBUFLEN) {

		if(data_pos >= data_end) {
			memset(driver->tmpbuf, 0, sizeof driver->tmpbuf);
			tmpcnt = GYMBUFLEN;
			break; /* f[^I[ (TODO:[v) */
		}

		op = *data_pos++;
		switch(op) {
		case 0: /* 1/60[sec]o ()*/

			/* 1/60[sec]~LVO܂B */
			ASSERT(tmpcnt + GYMBUFLEN_60 <= COUNTOF(driver->tmpbuf));
			ym2612_mix(ym2612, &driver->tmpbuf[0][tmpcnt], GYMBUFLEN_60);
			sn76489_mix(sn76489, &driver->tmpbuf[1][tmpcnt], GYMBUFLEN_60);
			gym_dacmix(&driver->tmpbuf[2][tmpcnt], dacbuf, daccnt);
			tmpcnt += GYMBUFLEN_60;

			break;

		case 1: /* YM2612 Part Io (=WX^ԍ, WX^l) */

			reg  = *data_pos++;
			data = *data_pos++;
			if((reg == 0x2a) && (daccnt < GYMBUFLEN_DAC)) {
				/* * ̂߁AYM2612 EmulatorDAC@\͎g܂B
				 *   DACo͂肵ċL^AGYM Driverɂă~LVO܂B
				 * * DAC en(0x2b)̐ݒ͉肹YM2612 Emulator֓n܂B
				 *   DACgṕAYM2612Ch6o͂OffɂKv邩łB
				 */
				dacbuf[daccnt++] = (data - 128) << 7/*DACʁA*/;
			} else {
				ym2612_write_reg(ym2612, 0, reg, data);
			}

			break;

		case 2: /* YM2612 Part IIo (=WX^ԍ, WX^l) */

			reg  = *data_pos++;
			data = *data_pos++;
			ym2612_write_reg(ym2612, 1, reg, data);

			break;

		case 3: /* SN76489o (=WX^l) */

			data = *data_pos++;
			sn76489_write(sn76489, (unsigned char)data);

			break;

		default: /* f[^G[? Ƃ肠 */

			break;
		}
	}

	/* ꎞobt@o̓obt@փRs[܂B */
	gym_copy(driver->tmpbuf[0], driver->tmpbuf[1], driver->tmpbuf[2], wbuff);

	/* Rs[Aꎞobt@O֋l߂܂B */
	tmpcnt -= GYMBUFLEN;
	memmove(driver->tmpbuf[0], &driver->tmpbuf[0][GYMBUFLEN], sizeof(short) * tmpcnt);
	memmove(driver->tmpbuf[1], &driver->tmpbuf[1][GYMBUFLEN], sizeof(short) * tmpcnt);
	memmove(driver->tmpbuf[2], &driver->tmpbuf[2][GYMBUFLEN], sizeof(short) * tmpcnt);

	driver->data_pos = data_pos;	/* ߂B */
	driver->data_end = data_end;	/* ߂B */
	driver->tmpcnt = tmpcnt;	/* ߂B */

	return 0;
}

int
gym_stream_callback(short* wbuff, int param)
{
	return gym_mix((GYMDRIVER*)param, wbuff);
}

/****************************************************************************
 *	AvP[Vp֐
 ****************************************************************************/

int
gym_play(const void* romp, int size)
{
	static GYMDRIVER driver; /* rs`shbłI */

	/* ܂mɒ~܂B */
	gym_stop();

	/* hCo܂B */
	if(gym_init(&driver, romp, size) != 0) {
		return -1;
	}

	/* Xg[ĐJn܂B */
	stream_play(GYMBUFLEN, gym_stream_callback, (int)&driver, 1/*X^bN؊gp*/);

	return 0;
}

void
gym_stop()
{
	/* Xg[Đ~܂B */
	stream_stop();

	/* hCõN[Abv͕svłB */
}
